<?php
// +--------------------------------------------------------------------------+
// | PayPal Plugin - geeklog CMS                                             |
// +--------------------------------------------------------------------------+
// | BaseIPN.class.php                                                        |
// |                                                                          |
// | This file contains the BaseIPN class, it provides an interface to        |
// | deal with IPN transactions from paypal                                   |
// +--------------------------------------------------------------------------+
// |                                                                          |
// | Copyright (C) 2005-2006 by the following authors:                        |
// |                                                                          |
// | Authors: Vincent Furia     - vinny01 AT users DOT sourceforge DOT net    |
// +--------------------------------------------------------------------------+
// |                                                                          |
// | This program is free software; you can redistribute it and/or            |
// | modify it under the terms of the GNU General Public License              |
// | as published by the Free Software Foundation; either version 2           |
// | of the License, or (at your option) any later version.                   |
// |                                                                          |
// | This program is distributed in the hope that it will be useful,          |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of           |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            |
// | GNU General Public License for more details.                             |
// |                                                                          |
// | You should have received a copy of the GNU General Public License        |
// | along with this program; if not, write to the Free Software Foundation,  |
// | Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.          |
// |                                                                          |
// +--------------------------------------------------------------------------+

// this file can't be used on its own
if (!defined ('VERSION')) {
    die ('This file can not be used on its own.');
}

/**
 * This file contains the BaseIPN class, it provides an interface to deal with
 * IPN transactions from paypal
 *
 * @author Vincent Furia <vinny01 AT users DOT sourceforge DOT net>
 * @copyright Vincent Furia 2005 - 2006
 * @package paypal
 */

/**
 * This class provides an interface to deal with IPN transactions from paypal
 *
 * @package paypal
 */
class BaseIPN {

    /**
     * Verify the transaction by check Paypal's servers
     *
     * Validate transaction by posting data back to the paypal webserver.  The response from
     * paypal should include 'VERIFIED' on a line by itself.
     *
     * @param array $in Array containing POST variables of transaction
     * @return boolean true if result successfully validated, false otherwise
     */
    function Verify($in) {
        global $_CONF;

        // Default verification to false
        $verified = false;

        // read the post from PayPal system and add 'cmd'
        $req = 'cmd=_notify-validate';

        // re-encode the transaction variables to be verified
        foreach ($in as $key => $value) {
            $value = urlencode(stripslashes($value));
            $req .= "&$key=$value";
        }

        // post back to PayPal system to validate
        $header .= "POST /cgi-bin/webscr HTTP/1.0\r\n";
        $header .= "Content-Type: application/x-www-form-urlencoded\r\n";
        $header .= "Content-Length: " . strlen($req) . "\r\n\r\n";
        $fp = fsockopen ($_CONF['paypal']['paypalURL'], 80, $errno, $errstr, 30);

        if (!$fp) {
            $this->handleFailure(PAYPAL_FAILURE_UNKNOWN, "Could not connect to paypal URL for verification");
            return false;
        } else {
            fputs ($fp, $header . $req);
            while (!feof($fp)) {
                $res = fgets ($fp, 1024);
                // For Debug Only
                // COM_errorLog("PAYPAL-IPN: IPN result -- $res")
                if (strcmp ($res, 'VERIFIED') == 0) {
                    $verified = true;
                    break;
                } else if (strcmp ($res, 'INVALID') == 0) {
                    $verified = false;
                    break;
                }
            }
            fclose ($fp);
        }

        return $verified;
    }

    /**
     * Log an IPN
     *
     * Logs the incoming IPN (serialized) along with the time it arrived, the originating IP address
     * and whether or not it has been verified (caller specified).  Also inserts the txn_id
     * seperately for look-up purposes.
     *
     * @param array $in POST variables of IPN transaction
     * @param boolean $verified true if transaction has been verified, false otherwise
     */
    function Log($in, $verified = false) {
        global $_SERVER, $_TABLES;

        // Change $verified into format for database
        if ($verified) {
            $verified = 1;
        } else {
            $verified = 0;
        }

        // Log to database
        $sql = "INSERT INTO {$_TABLES['paypal.ipnlog']} SET ip_addr = '{$_SERVER['REMOTE_ADDR']}', "
             . "time = NOW(), verified = $verified, txn_id = '{$in['txn_id']}', "
             . "ipn_data = '" . serialize($in) . '\'';
        DB_query($sql);

        return DB_insertId();
    }

    /**
     * Returns true if the the supplied email address is amoung allowed receiver addresses
     *
     * @param string $receiver_email Email to verify
     * @return boolean true if valid, false otherwise
     */
    function isValidEmail($receiver_email) {
        global $_CONF;

        return in_array($receiver_email, $_CONF['paypal']['receiverEmailAddr']);
    }

    /**
     * Checks to make sure that the transaction id is unique to prevent double counting.
     *
     * @param string $txn_id transaction id to verify
     * @return boolean true if unique, false otherwise
     */
    function isUniqueTxnId($txn_id) {
        global $_TABLES;

        // Count purchases with txn_id, if > 0
        $count = DB_count($_TABLES['paypal.purchases'], 'txn_id', $txn_id);
        if ($count > 0) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Confirms that payment status is complete (not 'denied', 'failed', 'pending', etc.)
     *
     * @param string $payment_status payment status to verify
     * @return boolean true if complete, false otherwise
     */
    function isStatusCompleted($payment_status) {
        return ($payment_status == 'Completed');
    }

    /**
     * Checks if payment status is reversed or refunded (ie some sort of cancelation)
     *
     * @param string $payment_status payment status to check
     * @return boolean true if payment status is reversed or refunded, false otherwise
     */
    function isStatusReversed($payment_status) {
        return ($payment_status == 'Reversed' || $payment_status == 'Refunded');
    }

    /**
     * Checks to make sure provided funds are sufficient to cover the cost of the purchased item(s)
     *
     * @param array $ids ids of items to check
     * @param array $quantity number ordered (per item)
     * @param real $payment_gross total payment made in current transaction
     * @param string $currency Currency of funds in payment_gross
     * @return boolean true if funds are sufficient, false otherwise
     */
    function isSufficientFunds($ids, $quantity, $payment_gross, $currency) {
        global $_CONF, $_TABLES;

        // Check currency
        if (!(isset($_CONF['paypal']['currency']) && strcasecmp($_CONF['paypal']['currency'], $currency) == 0)) {
            return false;
        }

        // Create a list of ids from $ids
        $idlist = implode(',', $ids);

        // Create/execute query string
        $sql = "SELECT id, price FROM {$_TABLES['paypal.products']} WHERE id in ($idlist)";
        $res = DB_query($sql);

        // Create a price lookup table
        while ($A = DB_fetchArray($res)) {
            $cost[$A['id']] = $A['price'];
        }

        // calculate the total purchase price
        $total = 0;
        for ($i = 0; $i < count($ids); $i++) {
            $total += $cost[$ids[$i]] * $quantity[$i];
        }

        // Compare total price to gross payment
        if ($total <= $payment_gross) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Process an incoming IPN transaction
     *
     * Do the following:
     * <ol>
     * <li>verify IPN</li>
     * <li>Log IPN</li>
     * <li>Check that transaction is complete</li>
     * <li>Check that transaction is unique</li>
     * <li>Check for valid receiver email address</li>
     * <li>Process IPN</li>
     * </ol>
     *
     * @param array $in POST variables of transaction
     * @return boolean true if processing valid and completed, false otherwise
     */
    function Process($in) {
        if (!$this->Verify($in)) {
            $logId = $this->Log($in, false);
            $this->handleFailure(PAYPAL_FAILURE_VERIFY, "($logId) Verification failed");
            return false;
        } else {
            $logId = $this->Log($in, true);
        }

        if (!$this->isStatusCompleted($in['payment_status'])) {
            // Not logged since this probably isn't an error
            // $this->handleFailure(PAYPAL_FAILURE_COMPLETED, "($logId) Status not complete");
            return false;
        }

        if (!$this->isUniqueTxnId($in['txn_id'])) {
            $this->handleFailure(PAYPAL_FAILURE_UNIQUE, "($logId) Non-unique transaction id");
            return false;
        }

        if (!$this->isValidEmail($in['receiver_email'])) {
            $this->handleFailure(PAYPAL_FAILURE_EMAIL, "($logId) Invalid receiver email address");
            return false;
        }

        switch ($in['txn_type']) {
            // buy now, donate, smart logos
            case 'web_accept':  //usually buy now
            case 'send_money':  //usually donation/send money
                // Process Buy Now, ignore donations
                if (!empty($in['item_number'])) {
                    $ids = array($in['item_number']);
                    $quantity = array($in['quantity']);
                    if (isset($in['settle_amount'])) {
                        $payment_gross = $in['mc_gross'] * $in['exchange_rate'];
                        $currency      = $in['settle_currency'];
                    } else {
                        $payment_gross = $in['mc_gross'];
                        $currency      = $in['mc_currency'];
                    }
                    if ($this->isSufficientFunds($ids, $quantity, $payment_gross, $currency)) {
                        $this->handlePurchase($ids, $quantity, $in);
                    } else {
                        $this->handleFailure(PAYPAL_FAILURE_FUNDS, "($logId) Insufficient funds for purchase");
                        return false;
                    }
                } else {
                    $this->handleDonation($in);
                }
                break;

            // shopping cart
            case 'cart':
                $ids = array();
                $quantity = array();
                for ($i = 1; $i <= $in['num_cart_items']; $i++) {
                    $ids[] = $in["item_number$i"];
                    $quantity[] = $in["quantity$i"];
                }
                if (isset($in['settle_amount'])) {
                    $payment_gross = $in['mc_gross'] * $in['exchange_rate'];
                    $currency      = $in['settle_currency'];
                } else {
                    $payment_gross = $in['mc_gross'];
                    $currency      = $in['mc_currency'];
                }
                if ($this->isSufficientFunds($ids, $quantity, $payment_gross, $currency)) {
                    $this->handlePurchase($ids, $quantity, $in);
                } else {
                    $this->handleFailure(PAYPAL_FAILURE_FUNDS, "($logId) Insufficient/incorrect funds for purchase");
                    return false;
                }
                break;

            // other, unknown, unsupported
            default:
                $this->handleFailure(PAYPAL_FAILURE_UNKNOWN, "($logId) Unknown transaction type");
                return false;
        }

        COM_errorLog("PAYPAL-IPN: purchases success",1);
        return true;
    }

    /**
     * Add a record of the purchase to the DB
     *
     * @param array $products Product Id(s) of Product(s) purchased
     * @param array $quantity Quantity of products purchases
     * @param array $paypal_data IPN POST variables
     * @todo implemente physical item vs. download, reflected in 'status'
     */
    function handlePurchase($products, $quantity, $paypal_data) {
        global $_TABLES, $_CONF;

        // initialize file and names arrays
        $files = array();
        $names = array();

        // for each item purchased, record purchase in purchase table
        for ($i = 0; $i < count($products); $i++) {
            // grab relevant product data from product table to insert into purchase table.
            $sql = "SELECT name, expiration, download, file FROM {$_TABLES['paypal.products']} "
                 . "WHERE id = {$products[$i]}";
            $res = DB_query($sql);
            $A = DB_fetchArray($res);
            if ($A['download'] > 0) {
                $files[] = $_CONF['path'] . '/plugins/paypal/files/' . $A['file'];
            }
            $names[] = $A['name'];

            // Do record anonymous users in purchase table
            if ( is_numeric($paypal_data['custom']) && $paypal_data['custom'] > 0 ) {
                // Add the purchase to the paypal purchase table
                $sql = "INSERT INTO {$_TABLES['paypal.purchases']} SET product_id = {$products[$i]}, "
                     . "quantity = {$quantity[$i]}, user_id = {$paypal_data['custom']}, "
                     . "txn_id = '{$paypal_data['txn_id']}', "
                     . 'purchase_date = NOW(), status = \'complete\'';

                /**
                 * @todo implemente physical item vs. download, reflected in 'status'
                 */
                // if physical item (aka, must be shipped) status = 'pending', otherwise 'complete'
                //if ( $physical == 1 ) {
                //    $sql .= ", status = 'pending'";
                //} else {
                //    $sql .= ", status = 'complete'";
                //}

                // add an expiration date if appropriate
                if (is_numeric($A['expiration'])) {
                    $sql .= ", expiration = DATE_ADD(NOW(), INTERVAL {$A['expiration']} DAY)";
                }
                DB_query($sql);
            }
        }

        /**
         * @todo update the email to reflect quantities...
         */
        // Send the purcahser a confirmation email (if set to do so in config.php)
        if ( ( is_numeric($paypal_data['custom']) && $paypal_data['custom'] != 1 &&
               $_CONF['paypal']['purchase_email_user'] ) ||
             ( (!is_numeric($paypal_data['custom']) || $paypal_data['custom'] == 1) &&
               $_CONF['paypal']['purchase_email_anon'] )) {
            // setup templates
            $message = new Template($_CONF['path'] . 'plugins/paypal/templates');
            $message->set_file(array('subject' => 'purchase_email_subject.txt',
                                     'message' => 'purchase_email_message.txt' ));
            // site variables
            $message->set_var('site_url', $_CONF['site_url']);
            $message->set_var('site_name', $_CONF['site_name']);
            $message->set_var('site_url', $_CONF['site_url']);

            // comma seperated list of product names
            $message->set_var('products', implode(', ', $names));

            // paypal details
            $message->set_var('payment_gross', $paypal_data['payment_gross']);
            $message->set_var('tax', $paypal_data['tax']);
            $message->set_var('shipping', $paypal_data['mc_shipping']);
            $message->set_var('handling', $paypal_data['mc_handling']);
            $message->set_var('payment_date', $paypal_data['payment_date']);
            $message->set_var('payer_email', $paypal_data['payer_email']);
            $message->set_var('first_name', $paypal_data['first_name']);
            $message->set_var('last_name', $paypal_data['last_name']);

            // parse templates for subject/text
            $subject = trim($message->parse('output', 'subject'));
            $text    = $message->parse('output', 'message');

            // if specified to mail attachment, do so, otherwise skip attachment
            if ( (( is_numeric($paypal_data['custom']) && $paypal_data['custom'] != 1 &&
                    $_CONF['paypal']['purchase_email_user_attach'] ) ||
                  ( (!is_numeric($paypal_data['custom']) || $paypal_data['custom'] == 1) &&
                    $_CONF['paypal']['purchase_email_anon_attach'] )) &&
                  count($files) > 0  ) {
                paypal_mailAttachment($paypal_data['payer_email'], $subject, $text, $files,
                                      $_CONF['paypal']['receiverEmailAddr'][0]);
            } else {
                COM_mail($paypal_data['payer_email'], $subject, $text,
                         $_CONF['paypal']['receiverEmailAddr'][0]);
            }
        }
    }

    /**
     * Not yet implemented
     *
     * @todo Implement handleSubscription
     */
    function handleSubscription() {
		$this->handleFailure(PAYPAL_FAILURE_UNKNOWN, "Subscription not handled");
    }

    /**
     * Not yet implemented
     *
     * @todo Implement handleDonation
     */
    function handleDonation() {
		$this->handleFailure(PAYPAL_FAILURE_UNKNOWN, "Donation not handled");
    }

    /**
     * Handle what to do in the event of a purchase/IPN failure
     *
     * This method does some basic failure handling.  For anything more
     * advanced it is recommend you override this method in IPN.class.php.
     *
     * @param int $type Type of failure that occurred
     * @param string $msg Failure message
     */
    function handleFailure($type = PAYPAL_FAILURE_UNKNOWN, $msg = '') {
        // Log the failure to geeklog's error log
        COM_errorLog('PAYPAL-IPN: ' . $msg,1);
    }
}

?>